#-*- encoding: utf-8; grammar-ext: py; mode: python -*-

#=========================================================================
"""
  Copyright |(c)| 2015 `Matt Bogosian`_ (|@posita|_).

  .. |(c)| unicode:: u+a9
  .. _`Matt Bogosian`: mailto:mtb19@columbia.edu
  .. |@posita| replace:: **@posita**
  .. _`@posita`: https://github.com/posita

  Please see the accompanying ``LICENSE`` (or ``LICENSE.txt``) file for
  rights and restrictions governing use of this software. All rights not
  expressly waived or licensed are reserved. If such a file did not
  accompany this software, then please contact the author before viewing
  or using this software in any capacity.
"""
#=========================================================================

from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)
from builtins import * # pylint: disable=redefined-builtin,unused-wildcard-import,useless-suppression,wildcard-import
from future.builtins.disabled import * # pylint: disable=redefined-builtin,unused-wildcard-import,useless-suppression,wildcard-import

#---- Imports ------------------------------------------------------------

import logging
import os
import sys
import unittest
# import twisted
from twisted.internet import (
    base as t_base,
    defer as t_defer,
)

from txrc.logging import SILENT

#---- Constants ----------------------------------------------------------

__all__ = ()

_LOG_LVL = os.environ.get('_TXRC_LOG_LVL')
_LOG_LVL = SILENT if not _LOG_LVL else logging.getLevelName(_LOG_LVL)
_LOG_FMT = os.environ.get('_TXRC_LOG_FMT')

#---- Initialization -----------------------------------------------------

# Python 3.4 complains that assertRaisesRegexp is deprecated in favor of
# assertRaisesRegex, which Python 2.7's unittest doesn't have; this
# monkey patch fixes all that
if not hasattr(unittest.TestCase, 'assertRaisesRegex'):
    unittest.TestCase.assertRaisesRegex = unittest.TestCase.assertRaisesRegexp

# TODO: swap checks once <https://tm.tl/#8110> is fixed
if sys.version_info[0] <= 2:
# if sys.version_info[0] <= 2 \
#         or ( twisted.version.major, twisted.version.minor ) >= ( 15, 5 ):
    t_base.DelayedCall.debug = True

t_defer.setDebugging(True)
logging.basicConfig(format=_LOG_FMT)
logging.getLogger().setLevel(_LOG_LVL)



#-*- encoding: utf-8; grammar-ext: py; mode: python -*-

#=========================================================================
"""
  Copyright |(c)| 2015 `Matt Bogosian`_ (|@posita|_).

  .. |(c)| unicode:: u+a9
  .. _`Matt Bogosian`: mailto:mtb19@columbia.edu
  .. |@posita| replace:: **@posita**
  .. _`@posita`: https://github.com/posita

  Please see the accompanying ``LICENSE`` (or ``LICENSE.txt``) file for
  rights and restrictions governing use of this software. All rights not
  expressly waived or licensed are reserved. If such a file did not
  accompany this software, then please contact the author before viewing
  or using this software in any capacity.
"""
#=========================================================================

from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)
from builtins import * # pylint: disable=redefined-builtin,unused-wildcard-import,useless-suppression,wildcard-import
from future.builtins.disabled import * # pylint: disable=redefined-builtin,unused-wildcard-import,useless-suppression,wildcard-import

#---- Imports ------------------------------------------------------------

import functools
import io
import logging as _logging
import sys
from twisted.internet import defer as t_defer

#---- Constants ----------------------------------------------------------

__all__ = ()

SILENT = -(sys.maxsize) - 1

_IS_PY3 = sys.version_info >= ( 3, 0 )
_LOGGER = _logging.getLogger(__name__)
_LOGGER_TOP_LEVEL = _logging.getLogger()

#---- Functions ----------------------------------------------------------

#=========================================================================
def formattraceback(failure, *args, **kw):
    """
    Returns a traceback string for ``failure``.

    :param failure: the failure for which to build the traceback string

    :type failure: :class:`twisted.python.failure.Failure`

    :param args: passed to
        :meth:`twisted.python.failure.Failure.printTraceback`

    :param kw: passed to
        :meth:`twisted.python.failure.Failure.printTraceback`
    """
    # See <https://tm.tl/#8044>
    buf = io.StringIO() if _IS_PY3 else io.BytesIO()
    failure.printTraceback(buf, *args, **kw)
    tb_str = buf.getvalue() if _IS_PY3 else buf.getvalue().decode('ascii')

    return tb_str.strip()

#=========================================================================
def logerrback(failure, logger=_LOGGER_TOP_LEVEL, log_lvl=_logging.DEBUG, msg='Unhandled error:', handled=(), suppress_msg_on_handled=True, reraise_handled=False):
    """
    Generic errback function to log failures.

    :param failure: the failure to inspect

    :type failure: :class:`twisted.python.failure.Failure`

    :param logger: the logger to use

    :type logger: :class:`logging.Logger`

    :param Integral log_lvl: level passed to :meth:`logging.Logger.log`

    :param str msg: the message to log or `None` for no message

    :param sequence handled: a sequence of :exc:`Exception`s whose stack
        traces will *not* be logged

    :param bool reraise: if `True`, causes ``failure`` to be returned
        (passed through)

    :param bool suppress_msg_on_handled: if `True`, suppress logging of
        ``msg`` ``failure.value`` matches ``handled``

    :returns: ``failure`` if ``reraise`` is `True`, otherwise `None`
    """
    is_handled = failure.check(*handled)

    if not is_handled \
            or not suppress_msg_on_handled:
        if msg is not None:
            logger.log(log_lvl, msg)

        logger.log(log_lvl, formattraceback(failure))

    if not is_handled \
            or reraise_handled:
        return failure

#=========================================================================
def logerrbackdl(dl_res, logger=_LOGGER, log_lvl=_logging.DEBUG, msg=None, handled=(), suppress_msg_on_handled=False):
    """
    Generic errback function to log individual failures from
    :class:`twisted.internet.defer.DeferredList`s. Each failure is logged
    using :func:`logerrback`.

    :param iterable dl_res: the deferred list results to inspect

    :param logger: the logger to use

    :type logger: :class:`logging.Logger`

    :param log_lvl: passed to :func:`logerrback`

    :param msg: passed to :func:`logerrback`

    :param handled: passed to :func:`logerrback`

    :param suppress_msg_on_handled: passed to :func:`logerrback`

    :returns: ``dl_res``
    """
    if dl_res is not None:
        for success, res in dl_res:
            if not success:
                logerrback(res, logger, log_lvl, msg, handled, suppress_msg_on_handled, reraise_handled=False)

    return dl_res

#=========================================================================
def logunhandlederr(log_lvl, logger=_LOGGER):
    """
    Decorates a maybe-deferred callable (the decorated callable always
    returns a deferred) to log unhandled errors. Example usage:

        .. code-block:: python
            :linenos:

            @logunhandlederr(logging.DEBUG)
            def raiseorreturn(raise=True):
                if raise:
                    raise Exception

                return 'Success!'

            d = raiseorreturn()

    :param Integral log_lvl: level passed to :meth:`logging.Logger.log`

    :param logger: the logger to use

    :type logger: :class:`logging.Logger`

    :return: the decorated callable
    """
    def wrap(_call):
        def _logunhandlederr(*__args, **__kw):
            target_d = t_defer.maybeDeferred(_call, *__args, **__kw)
            target_d.addErrback(logerrback, logger, log_lvl, handled=( Exception, ))

            return target_d

        try:
            _logunhandlederr = functools.wraps(_call)(_logunhandlederr)
        except AttributeError:
            pass

        return _logunhandlederr

    return wrap



#-*- encoding: utf-8; grammar-ext: py; mode: python -*-

#=========================================================================
"""
  Copyright |(c)| 2015 `Matt Bogosian`_ (|@posita|_).

  .. |(c)| unicode:: u+a9
  .. _`Matt Bogosian`: mailto:mtb19@columbia.edu
  .. |@posita| replace:: **@posita**
  .. _`@posita`: https://github.com/posita

  Please see the accompanying ``LICENSE`` (or ``LICENSE.txt``) file for
  rights and restrictions governing use of this software. All rights not
  expressly waived or licensed are reserved. If such a file did not
  accompany this software, then please contact the author before viewing
  or using this software in any capacity.
"""
#=========================================================================

from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)
from builtins import * # pylint: disable=redefined-builtin,unused-wildcard-import,useless-suppression,wildcard-import
from future.builtins.disabled import * # pylint: disable=redefined-builtin,unused-wildcard-import,useless-suppression,wildcard-import

#---- Imports ------------------------------------------------------------

import logging as _logging

from .retry import * # pylint: disable=wildcard-import
from .version import __version__

#---- Constants ----------------------------------------------------------

__all__ = ()

LOGGER = _logging.getLogger(__name__)



#-*- encoding: utf-8; grammar-ext: py; mode: python -*-

#=========================================================================
"""
  Copyright |(c)| 2015 `Matt Bogosian`_ (|@posita|_).

  .. |(c)| unicode:: u+a9
  .. _`Matt Bogosian`: mailto:mtb19@columbia.edu
  .. |@posita| replace:: **@posita**
  .. _`@posita`: https://github.com/posita

  Please see the accompanying ``LICENSE`` (or ``LICENSE.txt``) file for
  rights and restrictions governing use of this software. All rights not
  expressly waived or licensed are reserved. If such a file did not
  accompany this software, then please contact the author before viewing
  or using this software in any capacity.
"""
#=========================================================================

from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)
from builtins import * # pylint: disable=redefined-builtin,unused-wildcard-import,useless-suppression,wildcard-import
from future.builtins.disabled import * # pylint: disable=redefined-builtin,unused-wildcard-import,useless-suppression,wildcard-import

#---- Imports ------------------------------------------------------------

import doctest
import logging
import os
from os import path
import re
import types

import txrc
import tests

#---- Constants ----------------------------------------------------------

__all__ = (
    'load_tests',
    'mkloadtests',
)

_PATH_RE = r'(' + re.escape(path.sep) + r'.*)?$'

_DOCTEST_ROOTS = (
    txrc,
    tests,
)

_LOGGER = logging.getLogger(__name__)

#---- Functions ----------------------------------------------------------

#=========================================================================
def load_tests(_, tests, __): # pylint: disable=redefined-outer-name
    """
    >>> True is not False
    True
    """
    in_len = tests.countTestCases()
    tests = _load_tests(_, tests, __)
    out_len = tests.countTestCases()
    # Test whether at least one doctest is found (which could be the
    # doctest for this function)
    assert in_len < out_len

    return tests

#=========================================================================
def mkloadtests(roots):
    """
    Creates the `load_tests` hook function to be called by
    `unittest.loadTestsFromModule` during unittest discovery to find and
    include all `doctest`s. To work with `setuptools`, this function must
    be in whatever module is passed to `setuptools.setup` via the
    ``test_suite`` argument.

    When iterated over, ``roots`` must provide either the root module or
    pairs of strings in the form ``( 'root_mod_path', 'root_mod_name' )``,
    where ``root_mod_path`` is the path to the root module and
    ``root_mod_name`` is the name of that module (as it would be used in
    an ``import`` statement).

    To use in a module:

    .. code-block:: python
        :linenos:

        import mymodule
        ROOTS = ( mymodule, ) # equivalent to ( ( mymodule.__file__, mymodule.__name__ ), )
        load_tests = mkloadtests(ROOTS)

    :param iterable roots: the root definitions from which to discover the
        doctests

    :note: I should *not* have to write this myself. *Grrr!*
    """
    def load_tests(_, tests, __): # pylint: disable=redefined-outer-name,unused-argument
        finder = doctest.DocTestFinder(exclude_empty=False)

        for root_mod in roots:
            if isinstance(root_mod, types.ModuleType):
                root_mod_path, root_mod_name = root_mod.__file__, root_mod.__name__
            else:
                root_mod_path, root_mod_name = root_mod

            if path.splitext(path.basename(root_mod_path))[0] == '__init__':
                root_mod_path = path.dirname(root_mod_path)

            if path.isfile(root_mod_path):
                root_mod_iter = ( ( path.dirname(root_mod_path), None, ( path.basename(root_mod_path), ) ), )
            else:
                root_mod_iter = os.walk(root_mod_path)

            for dir_name, _, file_names in root_mod_iter:
                if not re.match(re.escape(root_mod_path) + _PATH_RE, dir_name):
                    continue

                mod_name = dir_name[len(root_mod_path):].replace(path.sep, '.').strip('.')

                if mod_name:
                    mod_name = root_mod_name + '.' + mod_name
                else:
                    mod_name = root_mod_name

                for file_name in file_names:
                    if not file_name.endswith('.py'):
                        continue

                    if file_name == '__init__.py':
                        test_mod_name = mod_name
                    else:
                        test_mod_name = mod_name + '.' + path.splitext(file_name)[0]

                    try:
                        tests.addTest(doctest.DocTestSuite(test_mod_name, test_finder=finder))
                    except Exception as err: # pylint: disable=broad-except
                        _LOGGER.warning('unable to load doctests from %s (%s)', test_mod_name, err)

        return tests

    return load_tests

#=========================================================================
_load_tests = mkloadtests(_DOCTEST_ROOTS)

#---- Initialization -----------------------------------------------------

if __name__ == '__main__':
    from unittest import main
    main()



#-*-mode: python; encoding: utf-8; test-case-name: tests.testutil-*-

#=========================================================================
"""
  Copyright |(c)| 2015 `Matt Bogosian`_ (|@posita|_).

  .. |(c)| unicode:: u+a9
  .. _`Matt Bogosian`: mailto:mtb19@columbia.edu
  .. |@posita| replace:: **@posita**
  .. _`@posita`: https://github.com/posita

  Please see the accompanying ``LICENSE`` (or ``LICENSE.txt``) file for
  rights and restrictions governing use of this software. All rights not
  expressly waived or licensed are reserved. If such a file did not
  accompany this software, then please contact the author before viewing
  or using this software in any capacity.

  Portions of this code are adapted from `this blog post by Terry Jones
  <http://blogs.fluidinfo.com/terry/2009/11/12/twisted-code-for-retrying-function-calls/>`__.
  Per its author's terms, its use herein is permitted under the
  `CC0 1.0 License`_.

  Portions of this code are adapted from `this Gist
  <https://gist.github.com/theduderog/735556>`__. Per its authors' terms,
  its use herein is permitted under the `CC0 1.0 License`_.

  .. _`CC0 1.0 License`: https://creativecommons.org/publicdomain/zero/1.0/
"""
#=========================================================================

from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)
from builtins import * # pylint: disable=redefined-builtin,unused-wildcard-import,useless-suppression,wildcard-import
from future.builtins.disabled import * # pylint: disable=redefined-builtin,unused-wildcard-import,useless-suppression,wildcard-import

#---- Imports ------------------------------------------------------------

import functools
import logging
from twisted.internet import defer as t_defer
from twisted.internet import task as t_task
from zope import interface # pylint: disable=import-error

from .logging import (
    SILENT,
    formattraceback,
)

#---- Constants ----------------------------------------------------------

__all__ = (
    'RetryingCaller',
    'TimeoutError',
    'calltimeout',
    'calltimeoutexc',
    'deferredtimeout',
)

_LOGGER = logging.getLogger(__name__)

#---- Exceptions ---------------------------------------------------------

#=========================================================================
class TimeoutError(Exception):
    ""

    #---- Public properties ----------------------------------------------

    target_d = None

#---- Interfaces ---------------------------------------------------------

#=========================================================================
class IBackoffGeneratorFactory(interface.Interface):
    """
    Factory for creating a backoff generator for use with a
    :class:`RetryingCaller`.
    """
    # pylint: disable=no-method-argument,no-self-argument,useless-suppression

    #---- Hooks ----------------------------------------------------------

    def buildbackoffgenerator(retries):
        """
        Factory method.

        :param Integral retries: the number of retries the backoff
            generator should attempt before finishing; usually this is
            passed from a :class:`RetryingCaller`

        :returns: a generator yielding at most ``retries`` delays
            (measured in seconds)
        """

#=========================================================================
class IFailureInspector(interface.Interface):
    """
    Inspects a :class:`twisted.python.failure.Failure` and decides whether
    or not the call the generated it should be retried.
    """
    # pylint: disable=no-method-argument,no-self-argument,useless-suppression

    #---- Hooks ----------------------------------------------------------

    def shouldretry(failure):
        """
        Called by a :class:`RetryingCaller` after each call failure.

        :param failure: the failure raised from the underlying call

        :type failure: :class:`twisted.python.failure.Failure`

        :returns: a tuple in the format ``( Failure, bool )``, where the
            first item is the :class:`~twisted.python.failure.Failure`
            that should be raised if the number of retries has been
            exhausted, or if the second item is `True`
        """

#=========================================================================
class IFailureInspectorFactory(interface.Interface):
    """
    Factory for creating a :class:`FailureInspector` for use with a
    :class:`RetryingCaller`.
    """
    # pylint: disable=no-method-argument,no-self-argument,useless-suppression

    #---- Hooks ----------------------------------------------------------

    def buildfailureinspector():
        """
        Factory method.

        :returns: a :class:`IFailureInspector` provider
        """

#---- Classes ------------------------------------------------------------

#=========================================================================
class RetryingCaller(object):
    """
    Partial with the ability to retry the call on failure. Adapted from
    `Terry Jones's proposal
    <http://blogs.fluidinfo.com/terry/2009/11/12/twisted-code-for-retrying-function-calls/>`__.

    :param Integral retries: the number of times to retry (0 means call
        once with no retries)

    :param backoff_generator_factory: the factory used to generate the
        backoff delays between each retry attempt

    :type backoff_generator_factory: :class:`IBackoffGeneratorFactory`

    :param failure_inspector_factory: the factory used to generate the
        failure inspector used to determin if a retry should be attempted

    :type failure_inspector_factory: :class:`IFactoryInspectorFactory`

    :param reactor: the reactor to use; if `None`, then
        `twisted.internet.reactor` is used

    :type reactor: :class:`twisted.internet.interfaces.IReactorTime`
    """

    #---- Public inner classes -------------------------------------------

    @interface.implementer(IBackoffGeneratorFactory)
    class DefaultBackoffGeneratoryFactoryMixin(object):
        """
        Implements the default :class:`IBackoffGeneratorFactory` provider
        for a :class:`RetryingCaller` as a mix-in.
        """

        #---- Public constants -------------------------------------------

        log_lvl = SILENT

        #---- Constructor ------------------------------------------------

        def __init__(self, *_, **__): # pylint: disable=unused-argument
            super().__init__()

        #---- Public hooks -----------------------------------------------

        @classmethod
        def buildbackoffgenerator(cls, retries):
            """
            A :attr:`IBackoffGeneratorFactory:buildbackoffgenerator`
            provider that yields delays that start with 0.25 and double
            for each subsequent attempt up to ``retries`` times.
            """
            for delay in cls._basegenerator(retries):
                _LOGGER.log(cls.log_lvl, 'retrying in %0.3f seconds', delay)

                yield delay

        #---- Private static methods -------------------------------------

        @staticmethod
        def _basegenerator(retries):
            return ( min((1 << e) / 4, 32.0) for e in range(retries) )

    @interface.implementer(IFailureInspector)
    class DefaultFailureInspectorMixin(object):
        """
        Implements the default :class:`IFailureInspector` provider for a
        :class:`RetryingCaller` as a mix-in.
        """

        #---- Public constants -------------------------------------------

        log_lvl = SILENT

        #---- Constructor ------------------------------------------------

        def __init__(self, *_, **__): # pylint: disable=unused-argument
            super().__init__()

        #---- Public hooks -----------------------------------------------

        @classmethod
        def shouldretry(cls, failure):
            """
            A :attr:`IFailureInspector.shouldretry` provider that passes
            through the underlying failure, which is usually ``failure``,
            unless it is a :class:`twisted.internet.defer.FirstError`,
            in which case the ``subFailure`` attribute of
            :attr:`twisted.internet.defer.Failure.value` will be passed
            through. This method signals that the call should be retried
            unless the underlying
            :attr:`twisted.internet.defer.Failure.value` is a
            :exc:`twisted.internet.defer.CancelledError`.
            """
            if isinstance(failure.value, t_defer.FirstError):
                failure = failure.value.subFailure

            raise_right_now = isinstance(failure.value, ( t_defer.CancelledError, ))

            _LOGGER.log(cls.log_lvl, 'call failed')
            _LOGGER.log(cls.log_lvl, formattraceback(failure))

            return failure, raise_right_now

    @interface.implementer(IFailureInspectorFactory)
    class DefaultBehavior(DefaultBackoffGeneratoryFactoryMixin, DefaultFailureInspectorMixin):
        """
        Implements the default behaviors for a :class:`RetryingCaller`.
        """

        #---- Public hooks -----------------------------------------------

        def buildfailureinspector(self):
            return self

    #---- Private constants ----------------------------------------------

    _DEFAULT_BEHAVIOR = DefaultBehavior()

    #---- Constructor ----------------------------------------------------

    #=====================================================================
    def __init__(self, retries, backoff_generator_factory=_DEFAULT_BEHAVIOR, failure_inspector_factory=_DEFAULT_BEHAVIOR, reactor=None):
        self._retries = retries
        self._backoff_generator_factory = backoff_generator_factory
        self._failure_inspector_factory = failure_inspector_factory

        if reactor is None:
            from twisted.internet import reactor

        self._reactor = reactor

    #---- Public hook methods --------------------------------------------

    #=====================================================================
    def __call__(self, _call):
        """
        Allows a :class:`RetryingCaller` object to be used as a
        decorator:

        .. code-block:: python
            :linenos:

            retry = RetryingCaller(retries=3, ...)
            @retry
            def calltoretry(...):
                ...
        """
        def _retrywrapper(*__args, **__kw):
            return self.retry(_call, *__args, **__kw)

        try:
            _retrywrapper = functools.wraps(_call)(_retrywrapper)
        except AttributeError:
            pass

        return _retrywrapper

    #---- Public methods -------------------------------------------------

    #=====================================================================
    def retry(self, call, *args, **kw):
        """
        Retries ``call(*args, **kw)`` upon failure.

        :returns: a :class:`twisted.internet.defer.Deferred` for ``call``
            whose errback will be called with the most recent
            :class:`twisted.python.failure.Failure` returned by
            :meth:`FailureInspector.shouldretry` once all retries are
            exhausted
        """
        backoff_gen = self._backoff_generator_factory.buildbackoffgenerator(self._retries)
        failure_inspector = self._failure_inspector_factory.buildfailureinspector()

        def _retry(_failure=None):
            if _failure is not None:
                tested_failure, raise_right_now = failure_inspector.shouldretry(_failure)

                if not raise_right_now:
                    try:
                        delay = next(backoff_gen)
                    except StopIteration:
                        raise_right_now = True

                if raise_right_now:
                    return tested_failure
            else:
                delay = 0

            _d = t_task.deferLater(self._reactor, delay, call, *args, **kw)
            _d.addErrback(_retry)

            return _d

        return _retry()

#---- Functions ----------------------------------------------------------

#=========================================================================
def calltimeout(reactor, timeout, call, *args, **kw):
    """
    Shorthand for ``calltimeoutexc(reactor, timeout, call, None, *args,
    **kw)``.
    """
    return calltimeoutexc(reactor, timeout, call, None, *args, **kw)

#=========================================================================
def calltimeoutexc(reactor, timeout, call, timeout_exc, *args, **kw):
    """
    Calls :func:`twisted.internet.defer.maybeDeferred` on ``call``,
    ``args``, and ``kw`` and passes the result as the ``target_d``
    argument to :func:`deferredtimeout`.

    :param reactor: the reactor to use; if `None`, then
        `twisted.internet.reactor` is used

    :type reactor: :class:`twisted.internet.interfaces.IReactorTime`

    :param Integral timeout: the timeout in seconds; if ``timeout`` is
        less than zero, there is no timeout

    :param callable call: the callable

    :param Exception timeout_exc: the exception to raise instead of a
        :exc:`TimeoutError`

    :param args: passed to ``call``

    :param kw: passed to ``call``

    :returns: a :class:`twisted.internet.defer.Deferred` wrapping
        ``call``, ``args``, and ``kw``
    """
    return deferredtimeout(reactor, timeout, t_defer.maybeDeferred(call, *args, **kw), timeout_exc)

#=========================================================================
def deferredtimeout(reactor, timeout, target_d, timeout_exc=None):
    """
    Wraps ``target_d`` with a :class:`twisted.internet.defer.Deferred`
    that calls :meth:`~twisted.internet.defer.Deferred.cancel` on
    ``target_d`` and returns a :class:`twisted.python.failure.Failure`
    with a :exc:`TimeoutError` after ``timeout`` seconds if ``target_d``
    hasn't yet fired and ``timeout >= 0``.

    :param reactor: the reactor to use; if `None`, then
        `twisted.internet.reactor` is used

    :type reactor: :class:`twisted.internet.interfaces.IReactorTime`

    :param Integral timeout: the timeout in seconds; if ``timeout`` is
        less than zero, there is no timeout

    :param target_d: the target

    :type target_d: :class:`twisted.internet.defer.Deferred`

    :param Exception timeout_exc: the exception to raise instead of a
        :exc:`TimeoutError`

    :returns: a :class:`twisted.internet.defer.Deferred` that wraps
        ``target_d`` if ``timeout >= 0``, otherwise ``target_d``
    """
    if timeout < 0:
        return target_d

    if reactor is None:
        from twisted.internet import reactor

    def _timeout():
        if timeout_exc is None:
            exc = TimeoutError()
            exc.target_d = target_d
        else:
            exc = timeout_exc

        try:
            raise exc
        except Exception as exc: # pylint: disable=broad-except
            timeout_d.errback(exc)

        target_d.cancel()

    deadline = reactor.callLater(timeout, _timeout)

    def _canceler(_):
        if deadline.active():
            deadline.cancel()

        target_d.cancel()

    timeout_d = t_defer.Deferred(_canceler)

    def _handler(_passthru):
        if deadline.active():
            deadline.cancel()

        return _passthru

    timeout_d.addBoth(_handler)

    def _suppressalreadycalled(_failure):
        _failure.trap(t_defer.AlreadyCalledError)

    target_d.chainDeferred(timeout_d)
    target_d.addErrback(_suppressalreadycalled)

    return timeout_d



#!/usr/bin/env python
#-*- encoding: utf-8; grammar-ext: py; mode: python -*-

#=========================================================================
"""
  Copyright |(c)| 2015 `Matt Bogosian`_ (|@posita|_).

  .. |(c)| unicode:: u+a9
  .. _`Matt Bogosian`: mailto:mtb19@columbia.edu
  .. |@posita| replace:: **@posita**
  .. _`@posita`: https://github.com/posita

  Please see the accompanying ``LICENSE`` (or ``LICENSE.txt``) file for
  rights and restrictions governing use of this software. All rights not
  expressly waived or licensed are reserved. If such a file did not
  accompany this software, then please contact the author before viewing
  or using this software in any capacity.
"""
#=========================================================================

from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)
from builtins import * # pylint: disable=redefined-builtin,unused-wildcard-import,useless-suppression,wildcard-import
from future.builtins.disabled import * # pylint: disable=redefined-builtin,unused-wildcard-import,useless-suppression,wildcard-import

#---- Imports ------------------------------------------------------------

import logging
import sys
from twisted.internet import defer as t_defer
from twisted.internet import task as t_task
from twisted.trial import unittest as t_unittest

from txrc.retry import (
    RetryingCaller,
    TimeoutError,
    calltimeout,
    calltimeoutexc,
)
import tests # pylint: disable=unused-import

#---- Constants ----------------------------------------------------------

__all__ = ()

_LOGGER = logging.getLogger(__name__)

#---- Classes ------------------------------------------------------------

#=========================================================================
class CallTimeoutTestCase(t_unittest.TestCase):

    longMessage = True

    #---- Public hooks ---------------------------------------------------

    def setUp(self):
        super().setUp()
        self._clock = t_task.Clock()
        self._state = CallTimeoutTestCase._State(self._clock)

    def tearDown(self):
        super().tearDown()

    def test_deferred_cancel(self):
        delay = 2
        call_val = 'done'
        d = calltimeout(self._clock, delay, self._state.deferredcall, call_val, sys.maxsize >> 1)
        self._clock.advance(0)
        d.cancel()
        self._clock.advance(1)
        self. assertFailure(d, t_defer.CancelledError)
        self.assertFalse(self._state.fired)
        self.assertEqual(len(self._clock.getDelayedCalls()), 0)
        d.addErrback(lambda _res: None) # silence the unhandled error

    def test_deferred_delayed(self):
        delay = 2
        call_val = 'done'
        d = calltimeout(self._clock, delay, self._state.deferredcall, call_val)
        self._clock.advance(0)
        self.assertFalse(hasattr(d, 'result'))

        for _ in range(delay + 1):
            self._clock.advance(1)
            self.assertEqual(d.result, call_val)
            self.assertTrue(self._state.fired)

    def test_deferred_immediate(self):
        delay = 0
        call_val = 'done'
        d = calltimeout(self._clock, delay, self._state.deferredcall, call_val)
        self._clock.advance(0)
        self. assertFailure(d, TimeoutError)
        self.assertFalse(self._state.fired)
        self._clock.advance(2)
        self. assertFailure(d, TimeoutError)
        self.assertFalse(self._state.fired)
        self.assertFalse(self._clock.getDelayedCalls())
        d.addErrback(lambda _res: None) # silence the unhandled error

    def test_deferred_immediate_exc(self):
        delay = 0
        call_val = 'done'
        d = calltimeoutexc(self._clock, delay, self._state.deferredcall, ValueError, call_val)
        self._clock.advance(0)
        self.assertIsInstance(d.result.value, ValueError)
        self.assertFalse(self._state.fired)
        self.assertFalse(self._clock.getDelayedCalls())
        d.addErrback(lambda _res: None) # silence the unhandled error

    def test_deferred_negative_timeout(self):
        delay = -1
        call_val = 'done'
        d = calltimeout(self._clock, delay, self._state.deferredcall, call_val, 9)
        self._clock.advance(10)
        self.assertEqual(d.result, call_val)
        self.assertTrue(self._state.fired)
        self.assertFalse(self._clock.getDelayedCalls())

    def test_nondeferred_delayed(self):
        delay = 2
        call_val = 'done'
        d = calltimeout(self._clock, delay, self._state.nondeferredcall, call_val)
        self._clock.advance(0)
        self.assertEqual(d.result, call_val)
        self.assertTrue(self._state.fired)

        for _ in range(delay + 1):
            self._clock.advance(1)
            self.assertEqual(d.result, call_val)
            self.assertTrue(self._state.fired)

        self.assertFalse(self._clock.getDelayedCalls())

    def test_nondeferred_immediate(self):
        delay = 0
        call_val = 'done'
        d = calltimeout(self._clock, delay, self._state.nondeferredcall, call_val)
        self._clock.advance(0)
        self.assertEqual(d.result, call_val)
        self.assertTrue(self._state.fired)
        self.assertFalse(self._clock.getDelayedCalls())

    #---- Private inner classes ------------------------------------------

    class _State(object):

        #---- Constructor ------------------------------------------------

        def __init__(self, clock):
            self.clock = clock
            self.fired = False

        #---- Public methods ---------------------------------------------

        def deferredcall(self, val, wait_seconds=1):
            d = t_task.deferLater(self.clock, wait_seconds, self.nondeferredcall, val)

            return d

        def nondeferredcall(self, val):
            self.fired = True

            return val

#=========================================================================
class RetryingCallerTestCase(t_unittest.TestCase):

    longMessage = True

    #---- Public hooks ---------------------------------------------------

    def setUp(self):
        super().setUp()
        self._clock = t_task.Clock()

    def tearDown(self):
        super().tearDown()
        del self._clock

    def test_first_error(self):
        err_msg = 'Weee!'

        def _none(*_, **__): # pylint: disable=unused-argument
            return

        def _raise(*_, **__): # pylint: disable=unused-argument
            raise RuntimeError(err_msg)

        def _call():
            dl = t_defer.DeferredList(( t_defer.maybeDeferred(_none), t_defer.maybeDeferred(_raise) ), fireOnOneErrback=True, consumeErrors=True)

            return dl

        retrying_caller = RetryingCaller(0, reactor=self._clock)
        d = retrying_caller.retry(_call)
        self._clock.advance(0)
        self. assertFailure(d, RuntimeError)
        self.assertEqual(d.result.args[0], err_msg)
        d.addErrback(lambda _res: None) # silence the unhandled error

    def test_retry(self):
        retries = 5
        retrying_caller = RetryingCaller(retries, reactor=self._clock)

        def _mkflaky(_fail_this_many_times):
            def __calltoretry():
                _wrappedcalltoretry.times_called += 1

                if _wrappedcalltoretry.failures_left > 0:
                    _wrappedcalltoretry.failures_left -= 1
                    raise ValueError(_wrappedcalltoretry)

                return _wrappedcalltoretry

            _wrappedcalltoretry = retrying_caller(__calltoretry)
            _wrappedcalltoretry.times_called = 0
            _wrappedcalltoretry.failures_left = _fail_this_many_times

            return _wrappedcalltoretry

        def _checkdlresult(res):
            self.assertEqual(len(res), 3)
            self.assertTrue(res[0][0])
            self.assertEqual(res[0][1].times_called, retries)
            self.assertEqual(res[0][1].failures_left, 0)
            self.assertTrue(res[1][0])
            self.assertEqual(res[1][1].times_called, retries + 1)
            self.assertEqual(res[1][1].failures_left, 0)
            self.assertFalse(res[2][0])
            self.assertIsInstance(res[2][1].value, ValueError)
            self.assertEqual(res[2][1].value.args[0].times_called, retries + 1)
            self.assertEqual(res[2][1].value.args[0].failures_left, 0)

        def _handleresult(_call):
            def __attachresult(__passthru):
                _call.result = __passthru

                return __passthru

            return __attachresult

        call1 = _mkflaky(retries - 1) # should succeed in time
        d1 = call1()
        d1.addBoth(_handleresult(call1))

        call2 = _mkflaky(retries) # should succeed in time
        d2 = call2()
        d2.addBoth(_handleresult(call2))

        call3 = _mkflaky(retries + 1) # should not succeed in time
        d3 = call3()
        d3.addBoth(_handleresult(call3))

        dl = t_defer.DeferredList(( d1, d2, d3 ), consumeErrors=True)
        dl.addCallback(_checkdlresult)

        call4 = _mkflaky(retries) # will be canceled before completion
        d4 = call4()
        d4.addBoth(_handleresult(call4))
        d4.addErrback(lambda _res: None) # silence the unhandled error
        self._clock.advance(0)
        d4.cancel()

        delays = list(RetryingCaller.DefaultBehavior._basegenerator(retries)) # pylint: disable=protected-access
        self._clock.pump(delays)
        self.assertIsInstance(call4.result.value, t_defer.CancelledError)
        self.assertEqual(call4.times_called, 1)
        self.assertEqual(call4.failures_left, retries - 1)

#---- Initialization -----------------------------------------------------

if __name__ == '__main__':
    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)
    from unittest import main
    main()



#!/usr/bin/env python
#-*- encoding: utf-8; grammar-ext: py; mode: python -*-

#=========================================================================
"""
  Copyright |(c)| 2015 `Matt Bogosian`_ (|@posita|_).

  .. |(c)| unicode:: u+a9
  .. _`Matt Bogosian`: mailto:mtb19@columbia.edu
  .. |@posita| replace:: **@posita**
  .. _`@posita`: https://github.com/posita

  Please see the accompanying ``LICENSE`` (or ``LICENSE.txt``) file for
  rights and restrictions governing use of this software. All rights not
  expressly waived or licensed are reserved. If such a file did not
  accompany this software, then please contact the author before viewing
  or using this software in any capacity.
"""
#=========================================================================

from __future__ import (
    absolute_import, division, print_function,
    # See <https://bugs.python.org/setuptools/issue152>
    # unicode_literals,
)

#---- Imports ------------------------------------------------------------

import setuptools

# See this e-mail thread:
# <http://www.eby-sarna.com/pipermail/peak/2010-May/003348.html>
import logging # pylint: disable=unused-import
import multiprocessing # pylint: disable=unused-import

import inspect
import os
from os import path

#---- Constants ----------------------------------------------------------

__all__ = ()

INSTALL_REQUIRES = [
    'Twisted >= 15.3',
    'future',
]

_MY_DIR = path.dirname(inspect.getframeinfo(inspect.currentframe()).filename)

#---- Initialization -----------------------------------------------------

_namespace = {
    '_version_path': path.join(_MY_DIR, 'txrc', 'version.py'),
}

if path.isfile(_namespace['_version_path']):
    with open(_namespace['_version_path']) as _version_file:
        exec(compile(_version_file.read(), _namespace['_version_path'], 'exec'), _namespace, _namespace) # pylint: disable=exec-used

with open(path.join(_MY_DIR, 'README.rst')) as _readme_file:
    README = _readme_file.read()

__version__ = _namespace.get('__version__')
__release__ = _namespace.get('__release__', __version__)

_SETUP_ARGS = {
    'name'                : 'txrc',
    'version'             : __version__,
    'author'              : 'Matt Bogosian',
    'author_email'        : 'mtb19@columbia.edu',
    'url'                 : 'https://txrc.readthedocs.org/en/{}/'.format(__release__),
    'license'             : 'MIT License',
    'description'         : 'Twisted Retrying Caller Inspired by Terry Jones',
    'long_description'    : README,

    # From <https://pypi.python.org/pypi?%3Aaction=list_classifiers>
    'classifiers': (
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Development Status :: 4 - Beta',
        'Framework :: Twisted',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: System :: Networking',
    ),

    'packages'            : setuptools.find_packages(exclude = ( 'tests', )),
    'include_package_data': True,
    'install_requires'    : INSTALL_REQUIRES,
    'test_suite'          : 'tests',
}

if __name__ == '__main__':
    os.environ['COVERAGE_PROCESS_START'] = os.environ.get('COVERAGE_PROCESS_START', path.join(_MY_DIR, '.coveragerc'))
    setuptools.setup(**_SETUP_ARGS)
